"""
MIT License

Copyright (c) 2017 cgalleguillosm, AlessioNetti

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from time import clock as _clock, sleep as _sleep
from datetime import datetime
from abc import abstractmethod, ABC
from accasim.utils.reader_class import reader
from accasim.utils.misc import CONSTANT, watcher_demon, DEFAULT_SIMULATION, load_config, path_leaf, clean_results
from accasim.utils.visualization_class import scheduling_visualization
from accasim.base.event_class import event, event_mapper
from accasim.base.resource_manager_class import resources_class, resource_manager
from accasim.base.scheduler_class import scheduler_base
from accasim.base.event_class import job_factory
from threading import Thread, Event as THEvent
from os import getpid as _getpid
from psutil import Process as _Process
from _functools import reduce
import sys, os
import inspect
import pathlib
import asyncio

class simulator_base(ABC):

    def __init__(self, _resource_manager, _reader, _job_factory, _dispatcher, config_file=None, **kwargs):
        self.constants = CONSTANT()
        self.define_default_constants(config_file, **kwargs)
        self.real_init_time = datetime.now()
        assert(isinstance(_reader, reader))
        self.reader = _reader
        assert(isinstance(_resource_manager, resource_manager))
        self.resource_manager = _resource_manager
        assert(isinstance(_job_factory, job_factory))
        assert(self.check_request(_job_factory.attrs_names)), 'System resources must be included in Job Factory descrpition.'
        self.job_factory = _job_factory
        assert(isinstance(_dispatcher, scheduler_base))
        self.dispatcher = _dispatcher

        self.mapper = event_mapper(self.resource_manager)
        self.show_config()
        if self.constants.OVERWRITE_PREVIOUS:
            self.remove_previous()
        
    @abstractmethod
    def start_simulation(self):
        raise NotImplementedError('Must be implemented!')

    @abstractmethod
    def load_events(self):
        raise NotImplementedError('Must be implemented!')

    def check_request(self, attrs_names):
        _system_resources = self.resource_manager.resources.system_resource_types
        for _res in _system_resources:
            if not(_res in attrs_names):
                return False
        return True
    
    def generate_enviroment(self, config_path):
        config = load_config(config_path)
        resources = resources_class(**config)
        return resources.resource_manager()
    
    def define_filepaths(self, **kwargs):
        kwargs['WORKLOAD_FILENAME'] = path_leaf(kwargs['WORKLOAD_FILEPATH'])[1]
        if 'RESULTS_FOLDER_PATH' not in kwargs:
            script_path, script_name = path_leaf(inspect.stack()[-1].filename) 
            rfolder = kwargs.pop('RESULTS_FOLDER_NAME')
            kwargs['RESULTS_FOLDER_PATH'] = os.path.join(script_path, rfolder)
        self.create_folder(kwargs['RESULTS_FOLDER_PATH'])
        return kwargs
        
    def create_folder(self, path):
        if not os.path.exists(path):
            os.makedirs(path)
        return path
    
    def set_workload_input(self, workload_path, **kwargs):
        return reader(workload_path, **kwargs)
    
    def prepare_arguments(self, possible_arguments, arguments):
        return {k:v for k, v in arguments.items() if k in possible_arguments}
    
    def define_default_constants(self, config_filepath, **kwargs):
        config = DEFAULT_SIMULATION.parameters
        for k, v in config.items():
            if k not in kwargs:
                kwargs[k] = v
        if config_filepath:
            for k, v in load_config(config_filepath).items():
                kwargs[k] = v
        kwargs = self.define_filepaths(**kwargs)
        self.constants.load_constants(kwargs)
        
    def show_config(self, **kwargs):
        print('Initializing the simulator')
        print('Settings: ')
        print('\tSystem Configuration file: {}'.format(self.constants.SYS_CONFIG_FILEPATH))
        print('\tWorkload file: {}'.format(self.constants.WORKLOAD_FILEPATH))
        print('\tResults folder: {}{}.'.format(self.constants.RESULTS_FOLDER_PATH, ', Overwrite previous files' if self.constants.OVERWRITE_PREVIOUS else ''))
        print('\t\t ({}) Dispatching Plan Output. Prefix: {}'.format(self.on_off(self.constants.SCHEDULING_OUTPUT), self.constants.SCHED_PREFIX))
        print('\t\t ({}) Statistics Output. Prefix: {}'.format(self.on_off(self.constants.STATISTICS_OUTPUT), self.constants.STATISTICS_PREFIX))
        print('\t\t ({}) Dispatching Plan. Pretty Print Output. Prefix: {}'.format(self.on_off(self.constants.PPRINT_OUTPUT), self.constants.PPRINT_PREFIX))
        print('\t\t ({}) Benchmark Output. Prefix: {}'.format(self.on_off(self.constants.BENCHMARK_OUTPUT), self.constants.BENCHMARK_PREFIX))
        print('Ready to Start')
        
    def on_off(self, state):
        return 'ON' if state else 'OFF'
    
    def remove_previous(self):
        _wouts = [(self.constants.SCHEDULING_OUTPUT, self.constants.SCHED_PREFIX), (self.constants.STATISTICS_OUTPUT, self.constants.STATISTICS_PREFIX),
            (self.constants.PPRINT_OUTPUT, self.constants.PPRINT_PREFIX), (self.constants.BENCHMARK_OUTPUT, self.constants.BENCHMARK_PREFIX)]
        
        _paths = [os.path.join(self.constants.RESULTS_FOLDER_PATH, _prefix + self.constants.WORKLOAD_FILENAME) for state, _prefix in _wouts if state]            
        clean_results(*_paths)


class hpc_simulator(simulator_base):
    
    def __init__(self, sys_config, workload, _scheduler, _resource_manager=None, _reader=None, _job_factory=None, _simulator_config=None, overwrite_previous=True,
        scheduling_output=True, pprint_output=False, benchmark_output=False, statistics_output=True, show_statistics=True, **kwargs):
        
        kwargs['OVERWRITE_PREVIOUS'] = overwrite_previous
        kwargs['SYS_CONFIG_FILEPATH'] = sys_config
        kwargs['WORKLOAD_FILEPATH'] = workload
        kwargs['SCHEDULING_OUTPUT'] = scheduling_output
        kwargs['PPRINT_OUTPUT'] = pprint_output
        kwargs['BENCHMARK_OUTPUT'] = benchmark_output
        kwargs['STATISTICS_OUTPUT'] = statistics_output
        kwargs['SHOW_STATISTICS'] = show_statistics

        if not _resource_manager:
            _resource_manager = self.generate_enviroment(sys_config)
        if not _job_factory:
            _jf_arguments = ['job_class', 'job_attrs', 'job_mapper']
            args = self.prepare_arguments(_jf_arguments, kwargs)
            _job_factory = job_factory(_resource_manager, **args)
        if not _reader:
            _reader_arguments = ['max_lines']
            args = self.prepare_arguments(_reader_arguments, kwargs)
            _reader = self.set_workload_input(workload, **args)
        _scheduler.set_resource_manager(_resource_manager)
        
        simulator_base.__init__(self, _resource_manager, _reader, _job_factory, _scheduler, config_file=_simulator_config, **kwargs)
        
        self.start_time = None
        self.end_time = None
        self.max_sample = 2
        self.daemons = {}
        self.loaded_jobs = 0
               
    def monitor_datasource(self, _stop):
        '''
        runs continuously and updates the global data
        Useful for daemons
        '''
        while (not _stop.is_set()):
            self.constants.running_at['current_time'] = self.mapper.current_time
            self.constants.running_at['running_jobs'] = {x: self.mapper.events[x] for x in self.mapper.running}
            _sleep(self.constants.running_at['interval'])
    
    def start_simulation(self, init_unix_time, watcher=False, visualization=False, *args, **kwargs):
        # TODO Load dynamically as daemon_init.
        # The initial values could be set in the simulation call, but also the datasource for these variables could be setted in the call.
        # Obviously the monitor must load also dynamically.
        #======================================================================= 
        if visualization:
            running_at = {
                'interval': 1,
                'current_time': self.mapper.current_time,
                'running_jobs': {}
            }
            self.constants.load_constant('running_at', running_at)
            _stop = THEvent()
            monitor = Thread(target=self.monitor_datasource, args=[_stop])
            monitor.daemon = True
            self.daemons['visualization'] = {
                'class': scheduling_visualization,
                'args': [(None, 'constants.running_at'), (None, 'resource_manager.resources.system_capacity',)],
                'object': None
            }
            monitor.start()

            
        if watcher:
            functions = {
                'usage_function': self.mapper.usage,
                'availability_function': self.mapper.availability,
                'simulated_status_function': self.mapper.simulated_status,
                'current_time_function': self.mapper.simulated_current_time
            }
            self.daemons['watcher'] = {
                'class': watcher_demon,
                'args': [self.constants.WATCH_PORT, functions],
                'object': None
            }
        self.reader.open_file()

        if 'tweak_function' not in kwargs:
            kwargs['tweak_function'] = self.default_tweak_function
        kwargs['init_unix_time'] = init_unix_time

        simulation = Thread(target=self.start_hpc_simulation, args=args, kwargs=kwargs)
        simulation.start()

        # Starting the daemons
        self.daemon_init()
        simulation.join()
        # Stopping the daemons    
        [d['object'].stop() for d in self.daemons.values() if d['object']]
        if visualization:
            _stop.set()
        
    def default_tweak_function(self, _dict, init_unix_time):
        # At this point, where this function is called, the _dict object have the assignation from the log data.
        # As in the SWF workload logs the numbers of cores are not expressed, just the number of requested processors, we have to tweak this information
        # i.e we replace the number of processors by the number of requested cores.
        _processors = _dict.pop('requested_number_processors') if _dict['requested_number_processors'] != -1 else _dict.pop(
            'allocated_processors')
        _memory = _dict.pop('requested_memory') if _dict['requested_memory'] != -1 else _dict.pop('used_memory')
        # For this system configuration. Each processor has 2 cores, then total required cores are calculated as
        # required_processor x 2 cores =  required cores of the job
        _dict['core'] = _processors * 2
        # The requested memory is given for each processor, therefore the total of requested memory is calculated as memory x processors = required memory of the job
        _dict['mem'] = _memory * _processors
        # If the following keys are not given, these are calculated by the job factory with the previous data
        # In this dataset there is no way to know how many cores of each node where requested, by default both cores are requested by processor
        # Each node has two processor, then it is possible to alocate upto 2 processor by node for the same job.
        _dict['requested_nodes'] = _processors
        _dict['requested_resources'] = {'core': 2, 'mem': _memory}
        # A final and important tweak is modifying the time corresponding to 0 time. For this workload as defined in the file
        # the 0 time corresponds to Sun Jul 28 09:04:05 CEST 2002 (1027839845 unix time)
        _dict['submit_time'] = _dict.pop('submit_time') + init_unix_time
        assert (
        _dict['core'] >= 0), 'Please consider to clean your data cannot exists requests with any info about core request.'
        assert (
        _dict['mem'] >= 0), 'Please consider to clean your data cannot exists requests with any info about mem request.'
        return _dict
        
    def start_hpc_simulation(self, _debug=False, tweak_function=None, init_unix_time=0):

        if tweak_function:
            assert(callable(tweak_function)), 'tweak_function argument must be a function.'
            self.tweak_function = tweak_function
            self.init_unix_time = init_unix_time        
        
        #=======================================================================
        # Load events corresponding at the "current time" and the next one
        #=======================================================================
        event_dict = self.mapper.events
        self.start_time = _clock()
        self.constants.load_constant('start_time', self.start_time)

        print('Starting the simulation process.')
        
        self.load_events(event_dict, self.mapper, _debug, self.max_sample)
        events = self.mapper.next_events()

        #=======================================================================
        # Loop until there are not loaded, queued and running jobs
        #=======================================================================
        while events or self.mapper.has_events():
            _actual_time = self.mapper.current_time
            
            benchStartTime = _clock() * 1000
            
            if _debug:
                print('{} INI: Loaded {}, Queued {}, Running {}, Finished {}'.format(_actual_time, len(self.mapper.loaded), len(self.mapper.queued), len(self.mapper.running), len(self.mapper.finished)))
            self.mapper.release_ended_events(event_dict)

            queuelen = len(events)
            schedStartTime = _clock() * 1000
            schedEndTime = schedStartTime
            if events:
                if _debug:
                    print('{} DUR: To Schedule {}'.format(_actual_time, len(events)))
                to_dispatch = self.dispatcher.schedule(self.mapper.current_time, event_dict, events, _debug)
                # to_dispatch = self.dispatcher.schedule(self.mapper.current_time, event_dict, events, len(self.mapper.finished) > 15000)
                if _debug:
                    print('{} DUR: To Dispatch {}. {}'.format(_actual_time, len(to_dispatch), self.resource_manager.resources.usage()))
                time_diff = 0
                schedEndTime = _clock() * 1000
                try:
                    self.mapper.dispatch_events(event_dict, to_dispatch, time_diff, _debug)
                except AssertionError as e:
                    print('{} DUR: {}'.format(_actual_time, e))
                    print('{} DUR: Loaded {}, Queued {}, Running {}, Finished {}'.format(_actual_time, len(self.mapper.loaded), len(self.mapper.queued), len(self.mapper.running), len(self.mapper.finished)))
                    _exit()
                                   
            if _debug:
                print('{} END: Loaded {}, Queued {}, Running {}, Finished {}'.format(_actual_time, len(self.mapper.loaded), len(self.mapper.queued), len(self.mapper.running), len(self.mapper.finished)))

            #===================================================================
            # Loading next jobs
            #===================================================================
            if len(self.mapper.loaded) < 10:
                sample = self.max_sample if(len(self.mapper.loaded) < self.max_sample) else 2
                self.load_events(event_dict, self.mapper, _debug, sample)
            #===================================================================
            # Continue with next events            
            #===================================================================
            events = self.mapper.next_events()

            if self.constants.BENCHMARK_OUTPUT:
                benchEndTime = _clock() * 1000
                benchMemUsage = self.memory_usage_psutil()
                scheduleTime = schedEndTime - schedStartTime
                dispatchTime = benchEndTime - benchStartTime - scheduleTime
                
                asyncio.create_subprocess_exec(
                    self.write_to_benchmark(_actual_time, queuelen, benchEndTime - benchStartTime, scheduleTime, dispatchTime, benchMemUsage)
                )
                # self.write_to_benchmark(_actual_time, queuelen, benchEndTime - benchStartTime, scheduleTime, dispatchTime, benchMemUsage)

        self.end_time = _clock()
        assert(self.loaded_jobs == len(self.mapper.finished)), 'Loaded {} and Finished {}'.format(self.loaded_jobs, len(self.mapper.finished))
        self.statics_write_out(self.constants.SHOW_STATISTICS, self.constants.STATISTICS_OUTPUT)
        print('Simulation process completed.')
        self.mapper.current_time = None

    def statics_write_out(self, show, save):
        if not(show or save):
            return
        wtimes = self.mapper.wtimes
        slds = self.mapper.slowdowns
        sim_time_ = 'Simulation time: {0:.2f} secs\n'.format(self.end_time - self.start_time)
        disp_method_ = 'Dispathing method: {}\n'.format(self.dispatcher)
        total_jobs_ = 'Total jobs: {}\n'.format(self.loaded_jobs)
        makespan_ = 'Makespan: {}\n'.format(self.mapper.last_run_time - self.mapper.first_time_dispatch)
        avg_wtimes_ = 'Avg. waiting times: {}\n'.format(reduce(lambda x, y:x + y, wtimes) / float(len(wtimes)))
        avg_slowdown_ = 'Avg. slowdown: {}\n'.format(reduce(lambda x, y:x + y, slds) / float(len(slds)))
        if show:
            print('\n\t' + '\t'.join([sim_time_, disp_method_, total_jobs_, makespan_, avg_wtimes_, avg_slowdown_]))
        if save:
            _filepath = os.path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.STATISTICS_PREFIX + self.constants.WORKLOAD_FILENAME)
            with open(_filepath, 'a') as f:
                f.write(sim_time_)            
                f.write(disp_method_)
                f.write(total_jobs_)
                f.write(makespan_)
                f.write(avg_wtimes_)
                f.write(avg_slowdown_)

    def write_to_benchmark(self, time, queueSize, stepTime, schedTime, simTime, memUsage):
        """
            Writes to an output file the resource usage string related to the current simulation step.

            The output string contains 6 fields, corresponding to the input arguments, separated by ":", and can
            be easily parsed by any Python program.

        :param time: the timestamp relative to the simulation step
        :param queueSize: the size of the queue at the simulation step (before scheduling)
        :param stepTime: the total time required to perform the simulation step
        :param schedTime: the time related to the scheduling procedure
        :param simTime: the remaining time used in the step, related to the simulation process
        :param memUsage: memory usage (expressed in MB) at the simulation step
        """
        bvalues = [time, queueSize, stepTime, schedTime, simTime, memUsage]
        sep_token = ';' 
        bline = sep_token.join([str(v) for v in bvalues]) + '\n'
        
        _filepath = os.path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.BENCHMARK_PREFIX + self.constants.WORKLOAD_FILENAME)
        with open(_filepath, 'a') as f:
            f.write(bline)

    def load_events(self, jobs_dict, mapper, _debug=False, time_samples=2):
        _time = None
        while not self.reader.EOF and time_samples > 0:
            _dicts = self.reader.next_dicts()
            if not _dicts:
                break
            tmp_dict = {}
            job_list = []
            for _dict in _dicts:
                if self.tweak_function:
                    self.tweak_function(_dict, self.init_unix_time)
                je = self.job_factory.factory(**_dict)
                if self.checkJobValidity(je):
                    self.loaded_jobs += 1
                    tmp_dict[je.id] = je
                    job_list.append(je)
                elif _debug:
                    print("Job %s violates the system's resource constraints and will be discarded" % je.id)
                if _time != je.queued_time:
                    _time = je.queued_time
                    time_samples -= 1
            mapper.load_events(job_list)
            jobs_dict.update(tmp_dict)

    def checkJobValidity(self, job):
        """
            Simple method that checks if the loaded job violates the system's resource constraints.

        :param job: Job object 
        :return: True if the job is valid, false otherwise
        """
        resGroups = self.resource_manager.groups_available_resource()
        validGroups = 0
        for group in resGroups.values():
            valid = True
            for k, res in job.requested_resources.items():
                if group[k] < res:
                    valid = False
                    break
            if valid:
                validGroups += 1
        return validGroups > 0

    def memory_usage_psutil(self):
        # return the memory usage in MB
        process = _Process(_getpid())
        memr = process.memory_info().rss / float(2 ** 20)
        return memr
    
    def daemon_init(self):         
        _iter_func = lambda act, next: act.get(next) if isinstance(act, dict) else (getattr(act, next)() if callable(getattr(act, next)) else getattr(act, next))
        for _name, d in self.daemons.items():
            _class = d['class']
            if not _class:
                continue
            _args = []
            for _arg in d['args']:
                if isinstance(_arg, tuple):
                    res = reduce(_iter_func, _arg[1].split('.'), self if not _arg[0] else _arg[0])
                    _args.append(res)
                else:
                    _args.append(_arg)
            self.daemons[_name]['object'] = _class(*_args)
            self.daemons[_name]['object'].start()



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
from copy import deepcopy
from accasim.utils.misc import CONSTANT, FrozenDict
import sys

class resources_class:
    """
        resources class: Stablish the resources, allocate and release their use.
        
    """
    ON = 1
    OFF = 0
    
    def __init__(self, groups, resources, **kwargs):
        """
            :param groups: define the groups of resources. i.e: {'group_0': {'core': 4, 'mem': 10}, .. }
            :param resources: Stablish the available resources of the system, in terms of number of previous groups. i.e: {'group_0': 32}, This will set 32 nodes of the group_0
            :param **kwargs:
                - node_prefix: This will set the prefix of the node name. The default name is 'node', this name is followed by _(i) where i corresponds to the ith loaded node.
                - available_prefix: This will set the prefix of the available resources. Internal use
                - used_prefix: This will set the prefix of the used resources. Internal use
        """
        self.constants = CONSTANT()
        self.groups = {}
        self.resources = {}
        self.resources_status = {}
        self.system_resource_types = []
        self.system_total_resources = None
        # self.resources_tree = resource_map(kwargs['groups'])
        # self.allocated = {}
        self.node_prefix = kwargs['node_prefix'] if 'node_prefix' in kwargs else 'node_' 
        self.available_prefix = kwargs['available_prefix'] if 'available_prefix' in kwargs else 'a_'
        self.used_prefix = kwargs['used_prefix'] if 'available_prefix' in kwargs else 'u_'

        for group_name, group_values in groups.items():
            self.system_resource_types += filter(lambda x: x not in self.system_resource_types, list(group_values.keys()))
            resource_group = { '%s%s' % (p, attr): q if p == self.available_prefix else 0
                for attr, q in group_values.items() for p in [self.available_prefix, self.used_prefix]
            }
            self.define_group(group_name, resource_group)

        j = 0
        for group_name, q in resources.items():
            for i in range(q):
                _node_name = '%s%i' % (self.node_prefix, j + 1)
                _attrs_values = self.groups[group_name]
                self.resources[_node_name] = deepcopy(_attrs_values)
                self.resources_status[_node_name] = self.ON
                # self.resources_tree.add(_node_name, kwargs['groups'][group_name])
                j += 1              

    def total_resources(self):
        if self.system_total_resources:
            return self.system_total_resources
        avl_types = {_type: 0 for _type in self.system_resource_types}
        #=======================================================================
        # for _node_values in self.resources.values():
        #     for _type in avl_types.keys():
        #         print(_type, self.available_resource_key(_type), _node_values)
        #         avl_types[_type] += _node_values[self.available_resource_key(_type)]
        #=======================================================================
        for _node, _node_values in self.resources.items():
            for _type in avl_types.keys():
                avl_types[_type] += _node_values[self.available_resource_key(_type)]
        self.system_total_resources = FrozenDict(**avl_types)
        return avl_types

    def define_group(self, name, group):
        assert(isinstance(group, dict))
        assert(name not in self.groups), ('Repreated name group: %s. Select another one.' % (name))
        self.groups[name] = group

    def allocate(self, node_name, **kwargs):
        # TODO: Update using self.system_resource_types
        assert(self.resources), 'The resources must be setted before jobs allocation'
        assert(self.resources_status[node_name] == self.ON), 'The Node {} is {}, it is impossible to allocate any job'
        _resources = self.resources[node_name]
        _used = {}
        for k, v in kwargs.items():
            _rem_attr = _resources['%s%s' % (self.available_prefix, k)] - _resources['%s%s' % (self.used_prefix, k)]
            assert(v <= _rem_attr), 'The event was request {} {}, but there is only {} available.'.format(v, k, _rem_attr)
            _resources['%s%s' % (self.used_prefix, k)] += v
            _used[k] = _rem_attr - v
        # self.resources_tree.update(node_name, _used)

    def release(self, node_name, **kwargs):
        # TODO: Update using self.system_resource_types
        assert(self.resources), 'The resources must be setted before release resources'
        assert(self.resources_status[node_name] == self.ON), 'The Node {} is {}.'
        _resources = self.resources[node_name]
        for k, v in kwargs.items():
            _resources['%s%s' % (self.used_prefix, k)] -= v
            assert(_resources['%s%s' % (self.used_prefix, k)] >= 0), 'The event was request to release %i %s, but there is only %i available. It is impossible less than 0 resources' % (v, k, _resources['%s%s' % (self.used_prefix, k)])
        #=======================================================================
        # self.resources_tree.update(node_name, {
        #     attr: (_resources['%s_%s' % (self.available_prefix, attr)] - _resources['%s_%s' % (self.used_prefix, attr)]) for attr in set([attr.split('_')[1] for attr in _resources.keys()])})
        #=======================================================================

    def availability(self):
        # TODO: Update using self.system_resource_types
        assert(self.resources)
        _a = {}
        for node, attrs in self.resources.items():
            if self.resources_status[node] == self.OFF:
                continue 
            _a[node] = {
                # attr: (attrs['%s%s' % (self.available_prefix, attr)] - attrs['%s%s' % (self.used_prefix, attr)]) for attr in set([attr.split('_')[1] for attr in attrs])
                attr: (attrs['%s%s' % (self.available_prefix, attr)] - attrs['%s%s' % (self.used_prefix, attr)]) for attr in self.system_resource_types
            }
        return _a

    def usage(self):
        # TODO: Update using self.system_resource_types
        _str = "System usage: "
        _str_usage = []
        usage = {k: 0 for k in list(self.resources.values())[0]}
        for attrs in self.resources.values():
            for k, v in attrs.items():
                usage[k] += v
        # for _attr in set([attr.split('_')[1] for attr in usage]):
        for _attr in self.system_resource_types:
            if usage['%s%s' % (self.available_prefix, _attr)] > 0:
                _str_usage.append("%s: %.2f%%" % (_attr, usage['%s%s' % (self.used_prefix, _attr)] / usage['%s%s' % (self.available_prefix, _attr)] * 100))
        return (_str + ', '.join(_str_usage))

    def system_capacity(self):
        _capacity = {
            r: {'total':
                sum([attrs[self.available_prefix + r] for _, attrs in self.resources.items()]) 
            } 
            for r in self.system_resource_types
        }
        return _capacity
    
    def resource_manager(self):
        return resource_manager(self)
    
    def available_resource_key(self, _key):
        assert(_key in self.system_resource_types), '{} is not a resource type'.format(_key)
        return '{}{}'.format(self.available_prefix, _key)        

    def __str__(self):
        _str = "Resources:\n"
        for node, attrs in self.resources.items():
            formatted_attrs = ""
            # for attr in set([attr.split('_')[1] for attr in attrs]):
            for attr in self.system_resource_types:
               formatted_attrs += '%s: %i/%i, ' % (attr, attrs['%s%s' % (self.used_prefix, attr)], attrs['%s%s' % (self.available_prefix, attr)])
            _str += '- %s: %s\n' % (node, formatted_attrs)
        return _str

class resource_manager:

    def __init__(self, _resource):
        assert(isinstance(_resource, resources_class)), ('Only %s class is acepted for resources' % resources_class.__name__)
        self.resources = _resource
        self.actual_events = {}

    def allocate_event(self, event, node_names):
        logging.debug('Allocating %s event in nodes %s' % (event.id, ', '.join([node for node in node_names])))
        _resources = event.requested_resources
        _attrs = event.requested_resources.keys()

        unique_nodes = [(t, node_names.count(t)) for t in set(node_names)]

        self.actual_events[event.id] = {
            node_name: { _attr:_resources[_attr] * q for _attr in _attrs} for (node_name, q) in unique_nodes
        }
        for node_name, values in self.actual_events[event.id].items():
            self.resources.allocate(node_name, **values)

    def remove_event(self, id):
        for node_name, values in self.actual_events.pop(id).items():
            self.resources.release(node_name, **values)

    def node_resources(self, *args):
        for arg in args:
            print(arg, self.resources.resources[arg])

    def availability(self):
        return self.resources.availability()

    def resource_types(self):
        return self.resources.system_resource_types

    def get_nodes(self):
        return list(self.resources.resources.keys())
    
    def get_total_resources(self, *args):
        """
            Return the total system resource for the required argument. The resource have to exist in the system. 
            If no arguments is proportioned all resources are returned.
            @param *args: Depends on the system configuration. But at least it must have ('core', 'mem') resources          
        """
        _resources = self.resources.total_resources()
        if not args or len(args) == 0:
            return {k: v for k, v in _resources.items()}
        avl_types = {}
        for arg in args:
            assert(arg in _resources), '{} is not a resource of the system. Available resource are {}'.format(arg, self.resource_types()) 
            avl_types[arg] = _resources[arg]
        return avl_types

    def groups_available_resource(self, _key=None):
        if not _key:
            _group = {}
            for k, v in self.resources.groups.items():
                _group[k] = {_type: v[self.resources.available_resource_key(_type)] for _type in self.resources.system_resource_types} 
            return _group
        _group_key = self.resources.available_resource_key(_key)
        return {_group:_v[_group_key]  for _group, _v in self.resources.groups.items()} 

    #===========================================================================
    # def get_used_resources(self):
    #     prf = 'u'
    #     used = {}
    #     for k in self.resource_types():
    #         s = '_'.join([prf, k])
    #         used[k] = sum([attrs[s] for attrs in self.resources.resources.values()])
    #     return used
    #===========================================================================



"""
MIT License

Copyright (c) 2017 cgalleguillosm, AlessioNetti

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import random
import sys
from abc import abstractmethod, ABC
from accasim.utils.misc import CONSTANT
from accasim.base.resource_manager_class import resource_manager



class allocator_base(ABC):
    """
    The base abstract interface all allocators must comply to.
    """

    def __init__(self, seed, resource_manager=None, **kwargs):
        """
        Allocator constructor (based on scheduler)

        :param seed: Seed if there is any random event
        :param res_man: resource manager for the system
        :param kwargs: Nothing for the moment         
        """
        random.seed(seed)
        self._constants = CONSTANT()
        self._avl_resources = []
        self._sorted_keys = []
        self.set_resource_manager(resource_manager)

    @abstractmethod
    def get_id(self):
        """
        Abstract method. Must be implemented by the subclass. 
        Must return the identification of the allocator. 
        
        :return: Allocator identification (for instance its name).    
        """
        raise NotImplementedError

    @abstractmethod
    def set_resources(self, res):
        """
        Abstract method. Must be implemented by the subclass.
        This method sets the internal reference to the dictionary of available resources in the system.
        If the reference points to a list used also outside of this class, the object should be deepcopied.
        
        If necessary, the resources are also sorted.
            
        :param res: the list of currently available resources in the system.       
        """
        raise NotImplementedError

    @abstractmethod
    def set_attr(self, **kwargs):
        """
        Abstract method. Must be implemented by the subclass.
        Method used to set internal parameters and meta-data for the allocator.
        
        Its behavior depends on the specific allocator that is being used, and some arguments may be discarded.
        
        :param kwargs: the internal parameters to be set, depending on the allocator
        """
        raise NotImplementedError

    @abstractmethod
    def allocating_method(self, es, cur_time, skip=False, reserved_time=None, reserved_nodes=None, debug=False):
        """
        Abstract method. Must be implemented by the subclass.
        This method must try to allocate the scheduled events contained in es. It will stop as soon as an event cannot
        be allocated, to avoid violations of the scheduler's priority rules, or proceed with other events depending
        on the skip parameter.
        
        The method must support both list of events for es, in which case it will return a list, or single events.
        If there is at least one successful allocation, avl_resources is updated and sorted again efficiently.

        :param es: the event(s) to be allocated
        :param cur_time: current time, needed to build the schedule list
        :param skip: determines if the allocator can skip jobs
        :param reserved_time: beginning of the next reservation slot (used for backfilling)
        :param reserved_nodes: nodes already reserved (used for backfilling)
        :param debug: Debugging flag

        :return: a list of assigned nodes of length e.requested_nodes, for all events that could be allocated. The
        list is in the format (time,event,nodes) where time can be either cur_time or None.
        """
        raise NotImplementedError
    
    def allocate(self, es, cur_time, skip=False, reserved_time=None, reserved_nodes=None, debug=False):
        """
        This is the method that is called by the Scheduler to allocate the scheduled jobs. First, It verifies the data consistency and availability, 
        and then call to the implemented allocation policy.   
        
        
        :param es: the event(s) to be allocated
        :param cur_time: current time, needed to build the schedule list
        :param skip: determines if the allocator can skip jobs
        :param reserved_time: beginning of the next reservation slot (used for backfilling)
        :param reserved_nodes: nodes already reserved (used for backfilling)
        :param debug: Debugging flag
        
        :return: the return of the implemented allocation policy.

        """
        assert(self.resource_manager is not None), 'The resource manager is not defined. It must defined prior to run the simulation.'
        if debug:
            print('{}: {} queued jobs to be considered in the dispatching plan'.format(cur_time, len(es)))
        return self.allocating_method(es, cur_time, debug)
    
    def set_resource_manager(self, _resource_manager):
        """
        Internally set the resource manager to deal with resource availability.
        
        :param _resource_manager: A resource manager instance or None. If a resource manager is already instantiated,
             it's used for set internally set it and obtain the system capacity for dealing with the request verifications.
             The dispathing process can't start without a resource manager. 
        """
        if _resource_manager:
            assert isinstance(_resource_manager, resource_manager), 'Resource Manager not valid for scheduler'
            self.resource_manager = _resource_manager
            self._base_availability = self.resource_manager.get_total_resources()
        else:
            self.resource_manager = None

    def __str__(self):
        """
            Retrieves the identification of the allocator.
        """
        return self.get_id()

class ffp_alloc(allocator_base):
    """
    A simple allocator. Does not sort the resources.
        
    This allocator supports both single events and lists of events. It also
     supports backfilling. No sorting of the resources is done, so they are
     considered as they are given in input.
    """

    name = 'First_Fit'

    def __init__(self, seed=0, resource_manager=None, **kwargs):
        """
        Constructor for the class.
        
        :param seed: seed for random events (not used)
        :param resource_manager: reference to the system resource manager
        :param kwargs: None at the moment
        """
        allocator_base.__init__(self, seed, resource_manager)
        if self.resource_manager:
            self._base_availability = self.resource_manager.get_total_resources()

    def get_id(self):
        return self.__class__.__name__

    def set_resources(self, res):
        """
        Sets in the internal variable avl_resources the current available resources for the system. It also sorts
        them, if the sort_resources method is implemented.
        
        :param res: the list of currently available resources for the system
        """
        self._avl_resources = res
        self._sorted_keys = self._sort_resources()

    def set_attr(self, **kwargs):
        """
        Method used to set internal parameters and meta-data for the allocator.

        Its behavior depends on the specific allocator that is being used, and some arguments may be discarded.
        It is not actively used in this simple allocator (for the moment).

        :param kwargs: None for the moment
        """
        pass

    def allocating_method(self, es, cur_time, skip=False, reserved_time=None, reserved_nodes=None, debug=False):
        """
        Given a job list es, this method searches for a suitable allocation for as many jobs as possible.
        
        In normal allocation, the method stops as soon as an event in the list cannot be allocated. In this case,
        ths list of already allocated jobs is returned. This is done to be coherent with the scheduler's rules.
        As an alternative, the skip parameter can be supplied to allow the scheduler to skip unallocated jobs.
        This method also support backfilling schedule. In this case, the backfilling parameters are supplied,
        and the allocator tries to fit jobs without delaying the reserved job. In this second case,
        the method does not stop when a job cannot be allocated, but simply skips it.
        
        es can be a list or a single event object. The return type (list or single tuple) changes accordingly.
        
        :param es: the event(s) to be allocated
        :param cur_time: current time, needed to build the schedule list
        :param skip: determines if the allocator can skip jobs
        :param reserved_time: beginning of the next reservation slot (used for backfilling)
        :param reserved_nodes: nodes already reserved (used for backfilling)

        :return: a list of assigned nodes of length e.requested_nodes, for all events that could be allocated. The
        list is in the format (time,event,nodes) where time can be either cur_time or None.
        """
        if not isinstance(es, (list, tuple)):
            listAsInput = False
            es = [es]
        else:
            listAsInput = True

        allocation = []
        success_counter = 0
        for e in es:
            requested_nodes = e.requested_nodes
            requested_resources = e.requested_resources
            # We verify that the job does not violate the system's resource constraints
            for t in requested_resources.keys():
                assert requested_resources[t] * requested_nodes <= self._base_availability[t], 'There are %i %s total resources in the system, requested %i by job %s' % (self._base_availability[t], t, requested_resources[t] * requested_nodes, e.id)

            # If the input arguments relative to backfilling are not supplied, the method operates in regular mode.
            # Otherwise, backfilling mode is enabled, allowing the allocator to skip jobs and consider the reservation.
            nodes_to_discard = self._compute_reservation_overlaps(e, cur_time, reserved_time, reserved_nodes, debug)
            backfilling_overlap = False if len(nodes_to_discard) == 0 else True

            assigned_nodes = []
            nodes_left = requested_nodes
            for node in self._sorted_keys:
                # The algorithm check whether the given node belongs to the list of reserved nodes, in backfilling.
                # If it does, the node is discarded.
                resources = self._avl_resources[node]
                if backfilling_overlap and node in nodes_to_discard:
                    continue
                # We compute the number of job units fitting in the current node, and update the assignment
                fits = int(self._event_fits_node(resources, requested_resources))
                if nodes_left <= fits:
                    assigned_nodes += [node] * nodes_left
                    nodes_left = 0
                else:
                    assigned_nodes += [node] * fits
                    nodes_left -= fits
                if nodes_left <= 0:
                    break

            # If, after analyzing all nodes, the allocation is still not complete, the partial allocation
            # is discarded.
            if nodes_left > 0:
                assigned_nodes = []
            assert not assigned_nodes or requested_nodes == len(assigned_nodes), 'Requested' + str(requested_nodes) + ' got ' + str(len(assigned_nodes))

            # If a correct allocation was found, we update the resources of the system, sort them again, and
            # add the allocation to the output list.
            if assigned_nodes:
                allocation.append((cur_time, e.id, assigned_nodes))
                self._update_resources(assigned_nodes, requested_resources)
                self._adjust_resources(self._sorted_keys)
                success_counter += 1
                if debug:
                    print('Allocation successful for event %s' % (e.id))
            # If no correct allocation could be found, two scenarios are possible: 1) normally, the allocator stops
            # here and returns the jobs allocated so far 2) if the skip parameter is enabled, the job is just
            # skipped, and we proceed with the remaining ones.
            else:
                if debug:
                    print('Allocation failed for event %s with %s nodes left' % (e.id, nodes_left))
                allocation.append((None, e.id, []))
                if not skip:
                    # if jobs cannot be skipped, at the first allocation fail all subsequent jobs fail too
                    for ev in es[(success_counter + 1):]:
                        allocation.append((None, ev.id, []))
                    if debug:
                        print('Cannot skip jobs, %s additional pending allocations failed' % (len(es) - success_counter - 1))
                    break
        if debug:
            print('There were %s successful allocations out of %s events' % (success_counter, len(es)))
        return allocation if listAsInput else allocation[0]

    def _compute_reservation_overlaps(self, e, cur_time, reserved_time, reserved_nodes, debug=False):
        """
        This method considers an event e, the current time, and a list of reservation start times with relative
        reserved nodes, and returns the list of reserved nodes that cannot be accessed by event e because of overlap.
        
        :param e: the event to be allocated
        :param cur_time: the current time
        :param reserved_time: the list (or single element) of reservation times
        :param reserved_nodes: the list of lists (or single list) of reserved nodes for each reservation
        :param debug: the debug flag
        :return: the list of nodes that cannot be used by event e
        """
        if reserved_time is None or reserved_nodes is None:
            return []
        else:
            if not isinstance(reserved_time, (list, tuple)):
                if cur_time + e.expected_duration > reserved_time:
                    if debug:
                        print('Backfill: Event %s is overlapping with reservation at time %s in backfilling mode' % (e.id, reserved_time))
                    return reserved_nodes
                else:
                    return []
            else:
                overlap_list = []
                for ind, evtime in enumerate(reserved_time):
                    if cur_time + e.expected_duration > evtime:
                        if debug:
                            print('Backfill: Event %s is overlapping with reservation at time %s in backfilling mode' % (e.id, evtime))
                        overlap_list += reserved_nodes[ind]
                return list(set(overlap_list))

    def _update_resources(self, reserved_nodes, requested_resources):
        """
        Updates the internal avl_resources list after a successful allocation.
        
        :param reserved_nodes: the list of nodes assigned to the allocated job
        :param requested_resources: the list of resources requested by the job per each node
        """
        for node in reserved_nodes:
            resource = self._avl_resources[node]
            for attr, v in requested_resources.items():
                assert resource[attr] - v >= 0, 'In node {}, the resource {} is going below to 0'.format(node, attr)
                resource[attr] -= v

    def _sort_resources(self):
        """
        Method which sorts the available resources dict. Not used in this class, but can be overridden by extended
        classes.

        :return: the sorted list of node keys (in this case, identical to the original)
        """
        return self._avl_resources.keys()

    def _adjust_resources(self, sorted_keys):
        """
        Method which restores the resources' sorting after a successful allocation. Not used in this class, has
        to be overridden.

        :param sorted_keys: the list of keys that needs to be adjusted
        :return: none
        """
        pass

    def _event_fits_node(self, resources, requested_resources):
        """
        Checks if the job with requested_resources fits the node with resources available. Returns the number
        of job units that fit the node
        
        :param resources: the node's available resources
        :param requested_resources: the job's requested resources
        :return: the number of job units fitting in the node
        """
        # min_availability is the number of job units fitting in the node. It is initialized at +infty,
        # since we must compute a minimum
        min_availability = sys.maxsize
        # if a job requests 0 resources, the request is deemed as not valid
        valid_request = False
        for k, v in requested_resources.items():
            # for each resource type, we compute the number of job units fitting for it, and refresh the minimum
            if v > 0 and min_availability > (resources[k] // v):
                valid_request = True
                min_availability = resources[k] // v
                # if the minimum reaches 0 (no fit) we break from the cycle
                if min_availability <= 0:
                    min_availability = 0
                    break
        if valid_request:
            return min_availability
        else:
            return 0
        
class consolidate_alloc(ffp_alloc):
    """
        Consolidate Allocator
    It is an allocator which sorts the nodes basing on the amount of free resources, trying to consolidate.
        
    The less the available resources, the higher the priority.
    The allocator is based on allocator_simple, changing only the sort and adjust methods.   
    """
    
    name = 'Consolidate'

    def __init__(self, seed=0, resource_manager=None, **kwargs):
        """
        Constructor for the class.

        :param seed: seed for random events (not used)
        :param resource_manager: reference to the system resource manager
        :param kwargs: None at the moment
        """
        ffp_alloc.__init__(self, seed, resource_manager)

        self.ranking = lambda x: sum(self._avl_resources[x].values())
        """
            Defines the ranking operator for sorting. Must use the self._avl_resources argument
            (the available resource dictionary). x represents a key.
        """

    def _sort_resources(self):
        """
        This method sorts the keys of the available resources dictionary, basing on the ranking operator.
        
        It is called after the resources are set in the allocator.
        
        :return: the list of sorted keys (node ids) for the resources
        """
        assert self._avl_resources is not None, 'The dictionary of available resources must be non-empty.'
        return sorted(self._avl_resources.keys(), key=self.ranking, reverse=False)

    def _adjust_resources(self, sorted_keys):
        """
        Adjusts the sorting of the resources after a successful allocation. 
        
        This method still uses python's sort method, because the Timsort implementation has O(n) complexity
        on mostly sorted data. Even with a custom implementation, the average case would cost O(n) at best.
        
        :param sorted_keys: the list of keys, almost sorted, that needs to be adjusted
        """
        assert self._avl_resources is not None, 'The dictionary of available resources must be non-empty.'
        assert sorted_keys is not None, 'The list of keys must be non-empty'
        sorted_keys.sort(key=self.ranking, reverse=False)



"""
MIT License

Copyright (c) 2017 AlessioNetti, cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import socket, json
import os
import argparse
from accasim.utils.misc import CONSTANT, load_config, watcher_daemon

# A simple launch script to contact the watcher demon related to a running test session, and obtain information about
# it. By default the connection is launched on localhost, but the user can choose an IP if the tests are running
# on a different machine.
if __name__ == '__main__':
    CONFIG_FOLDER = '../config/'
    ESSENTIALS_FILENAME = 'essentials.config'
    const = CONSTANT()
    const.load_constants(load_config(os.path.join(CONFIG_FOLDER, ESSENTIALS_FILENAME)))

    parser = argparse.ArgumentParser(description="Client For HPC Simulator Watcher Daemon")
    parser.add_argument("-usage", action="store_true", dest="usage", help="Request current virtual resource usage.")
    parser.add_argument("-progress", action="store_true", dest="progress", help="Request current local progress.")
    parser.add_argument("-all", action="store_true", dest="all", help="Request all previous data.")
    parser.add_argument("-ip", action="store", dest="hostip", default="localhost", type=str, help="IP of server machine.")

    args = parser.parse_args()

    # Remember that commands and responses must not be longer than MAX_LENGTH!
    command = ''
    if args.usage:
        command = 'usage'
    elif args.progress:
        command = 'progress'
    elif args.all:
        command = 'all'
    else:
        print('No commands specified. Quitting...')
        exit()

    client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client_socket.connect((args.hostip, const.WATCH_PORT))
    client_socket.sendall(json.dumps(command).encode())
    data = json.loads(client_socket.recv(watcher_demon.MAX_LENGTH).decode())

    if 'input_filepath' in data:
        print('- Current test instance: {}'.format(data['input_filepath']))
    if 'progress' in data:
        print('\t Completion percentage: {:.2f}%'.format(data['progress'] * 100))
    if 'actual_time' in data:
        print("\t Current simulated time : {}".format(data['actual_time']))
    if all(key in data for key in ['number_testrun_now', 'number_testruns']):
        print('\t Test instance on current file No. of: {} / {}'.format(data['number_testrun_now'], data['number_testruns']))
    if 'simulation_status' in data:
        print('\t {}'.format(data['simulation_status']))
    if 'usage' in data:
        print('\t {}'.format(data['usage']))
    if 'time' in data:
        print("\t Real elapsed time : {:.2f} secs".format(data['time']))
    #===========================================================================
    # if args.usage:
    #     print('- ' + data['usage'])
    # elif args.globalprogress:
    #     print('- Current test instance: %s' % data['input_filepath'])
    #     print('- Testing file No. of: %i / %i' % (data['number_testfile_now'], data['number_testfiles']))
    #     print('- Test instance on current file No. of: %i / %i' % (data['number_testrun_now'], data['number_testruns']))
    # elif args.localprogress:
    #     print('- Current test instance: %s' % data['input_filepath'])
    #     print('- Completion percentage: %i%%' % int(data['progress'] * 100))
    #     print('- Elapsed time: %i secs' % data['time'])
    # else:
    #     print('- Unknown response received: ')
    #     print(data)
    #===========================================================================

    client_socket.close()
    exit()



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import re
from abc import abstractmethod, ABC
from accasim.utils.misc import CONSTANT, default_swf_parse_config
import sys
from builtins import issubclass

class workload_parser_base(ABC):
    
    @abstractmethod
    def parse_line(self, line):
        pass
    
class default_workload_parser(workload_parser_base):
    def __init__(self):
        workload_parser_base.__init__(self)
        """
            workloader_paser is a general class for parsing workload files.
            @param reg_exp: Dictionary where the name of the group is the key and the value
                its regular expresion. It must contain all the regular expresions in a sorted way,
                for appending one at the end of the previous and then recover the value(s) for each group.
            @param avoid_token: List of reg_exp to avoid reading lines. The lines that are avoided 
                won't be readed by the parser. 
        """
        self.reg_exp_dict, self.avoid_tokens = default_swf_parse_config
        #=======================================================================
        # assert(isinstance(reg_exp_dict, dict)), 'The regular expressions must be passed as dictionary. Groupname: Regular Exp'
        # self.reg_exp_dict = reg_exp_dict
        # assert(isinstance(avoid_tokens, (list, tuple))), 'The tokens to avoid (also regular expressions) must be passed in a list'
        # self.avoid_tokens = avoid_tokens
        #=======================================================================
        
    def feasible_line(self, line):
        """
            @param line: Line to be checked
            @return: True if it can be parse (it does not match to any avoid token), False otherwise.
        """
        for _token in self.avoid_tokens:
            p = re.compile(_token)
            if p.fullmatch(line):
                return False
        return True        
    
    def parse_line(self, line):
        """
            Parse a feasible line, returning a dict for all groupnames
            @param line: Line to be checked
        """
        if not self.feasible_line(line):
            return None
        reg_exp = r''
        for _key, _reg_exp in self.reg_exp_dict.items():
            reg_exp += _reg_exp[0].format(_key)
        p = re.compile(reg_exp)
        _matches = p.match(line)
        if not _matches:
            return None
        return {k:self.reg_exp_dict[k][1](v) for k, v in _matches.groupdict().items()}

class reader:
    def __init__(self, filepath, parser=None, max_lines=None):
        self.const = CONSTANT()
        if parser:
            if not isinstance(parser, workload_parser_base):
                assert(issubclass(parser, workload_parser_base)), 'Only default_workload_parser class can be used as parsers'
                self.parser = parser()
            else:
                assert(isinstance(parser, workload_parser_base)), 'Only default_workload_parser object can be used as parsers'
                self.parser = parser                
        else:
            self.parser = default_workload_parser()
        self.last_line = 0
        self.max_lines = max_lines
        self.filepath = filepath
        self.file = None
        self.EOF = True
        
    def __del__(self):
        if self.file:
            self.file.close()
            self.file = None   
            self.EOF = True 
        
    def open_file(self): 
        if self.file is None:
            self.file = open(self.filepath)
            self.EOF = False    
        return self.file
    
    def read_next_lines(self, n_lines=1):
        # logging.debug('Reading next %i lines.' % (n_lines))
        if not self.EOF:
            tmp_lines = 0
            lines = [] 
            for line in self.file:
                lines.append(line[:-1])
                self.last_line += 1
                tmp_lines += 1
                if tmp_lines == n_lines or (self.max_lines and self.max_lines == self.last_line):
                    break
            if tmp_lines < n_lines or (self.max_lines and self.max_lines == self.last_line):
                self.EOF = True
            if self.EOF and tmp_lines == 0:
                return None 
            return lines
        return None
    
    def next_dicts(self, n_lines=1):
        """
            @param n_lines: number of lines to be parsed
            @return: return a list of dicts corresponding to the lines   
        """    
        _dicts = []
        while len(_dicts) < n_lines:
            line = self.read_next_lines()
            # No more lines. End of File
            if not line:
                break
            _dict = self.parser.parse_line(line[0])
            if _dict:
                _dicts.append(_dict)
            elif self.max_lines:
                if self.max_lines == self.last_line:
                    self.EOF = False
                self.max_lines += 1
        return None if not _dicts else _dicts 


"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
import re
import sys, os
import time
from abc import abstractmethod, ABC
from pydoc import locate
import asyncio
from builtins import str, filter
from inspect import signature
from _functools import reduce
from types import MethodType
from accasim.utils.misc import CONSTANT, sorted_list, default_swf_mapper
from accasim.base.resource_manager_class import resource_manager


class attribute_type:
    
    def __init__(self, name, type_class=None, optional=False):
        """
        @param name: Attribute name
        @param type_class: Class type of attribute (str, int, float, etc.) for casting. If value  is already casted it is not necesary.
        @param optional: False by default. If it is True, the default value will be None and it is not required to give any value to this.  
        """
        assert(isinstance(name, str))
        self.name = name
        self.type = type_class
        self.optional = optional  
                
class event(ABC):
    
    def __init__(self, job_id, queued_time, duration):
        self.constants = CONSTANT()
        self.id = job_id  # uuid.uuid1()
        self.queued_time = queued_time
        self.start_time = None
        self.end_time = None
        self.duration = duration
        self.end_order = 0
    
    def subattr(self, obj, attrs):
        if isinstance(attrs, tuple):
            values = []
            for attr in list(attrs):
                values.append(self.subattr(obj, attr))
            return values
        sp_attr = attrs.split('.')
        if len(sp_attr) > 1:
            tmp = getattr(obj, sp_attr[0])
            return self.subattr(tmp, ''.join(sp_attr[1:]))
        return getattr(obj, sp_attr[0])
    
    def schd_write_out(self):
        if not self.constants.SCHEDULING_OUTPUT:
            return 
        _dict = self.constants.SCHEDULE_OUTPUT
        _attrs = {}
        for a, av in _dict['attributes'].items():
            _attrs[a] = locate(av[-1])(*self.subattr(self, av[:-1])) 
        output_format = _dict['format']
        format_elements = re.findall('\{(\w+)\}', output_format)
        values = {k: v for k, v in _attrs.items() if k in format_elements}
        _filepath = os.path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.SCHED_PREFIX + self.constants.WORKLOAD_FILENAME)
        with open(_filepath, 'a') as f:
            f.write(output_format.format(**values) + '\n')
    
    def schd_pprint_write_out(self):
        if not self.constants.PPRINT_OUTPUT:
            return 
        _dict = self.constants.PPRINT_SCHEDULE_OUTPUT
        _order = _dict['order']
        _attrs = {a: locate(av[-1])(*self.subattr(self, av[:-1])) for a, av in _dict['attributes'].items() if a in _order}
        output_format = _dict['format']
        format_elements = re.findall('\{(\w+)\}', output_format)
        values = [_attrs[k] for k in _order]
        _filepath = os.path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.PPRINT_PREFIX + self.constants.WORKLOAD_FILENAME)
        with open(_filepath, 'a') as f:
            if self.end_order == 1:
                f.write(output_format.format(*_order) + '\n')
            f.write(output_format.format(*values) + '\n')
                        
class job_factory:
    def __init__(self, _resource_manager, job_class=event, job_attrs=[], job_mapper=default_swf_mapper):
        """
        @param _resource_manager: The resource manager of the simulator. It is required for creating the job requests.
        @param job_class: The class to be created by the Factory. By default it uses the Event class, but any subclass of it can be used (modified versions). 
        @param job_attrs: The extra attributes (attribute_type class) (already job_id, queued_time and duration are mandatory) to be set in the JobEvent class
        @param job_mapper: Rename the the old key to a new key (using the value of the mapper dictionary)
        """
        assert(isinstance(_resource_manager, resource_manager))
        assert(issubclass(job_class, event)), 'Only subclasses of event class are accepted. Received: {} class'.format(_class.__name__)
        if job_attrs:
            assert(isinstance(job_attrs, list) and all(isinstance(attr_type, attribute_type) for attr_type in job_attrs))
        else:
            attrs = self.default_job_description()
        # self.resource_manager = _resource_manager
        self.group_resources = _resource_manager.groups_available_resource()
        self.system_resources = _resource_manager.resources.system_resource_types
        self.obj_type = job_class
        self.obj_parameters = list(signature(self.obj_type).parameters)
        self.attrs_names = []
        self.mandatory_attrs = {}
        self.optional_attrs = {}
        self.mapper = job_mapper
        
        for attr in attrs:
            _attr_name = attr.name
            assert(_attr_name not in self.attrs_names + self.obj_parameters), '{} attribute name already set. Names must be unique'.format(_attr_name)
            if attr.optional:
                assert(_attr_name in self.obj_parameters), '{} attribute name is mandatory.'.format(_attr_name)
                self.optional_attrs[_attr_name] = attr
            else:
                self.mandatory_attrs[_attr_name] = attr
            self.attrs_names.append(_attr_name)
                 
    def factory(self, **kwargs):
        for _old, _new in self.mapper.items():
            value = kwargs.pop(_old)
            kwargs[_new] = value
        _missing = list(filter(lambda x:x not in kwargs, set(self.obj_parameters + list(self.mandatory_attrs))))
        assert(not _missing), 'Missing attributes: {}'.format(', '.join(_missing))
        _tmp = self.obj_type(**{k:kwargs[k] for k in self.obj_parameters})
        setattr(_tmp, '_dict', kwargs)
        self.add_attrs(_tmp, self.mandatory_attrs, kwargs)
        self.add_attrs(_tmp, self.optional_attrs, kwargs)
        self.add_request(_tmp)
        return _tmp
        
    def add_attrs(self, obj, reference, values):
        for _attr in reference:
            _type = reference[_attr].type
            _value = None
            if not reference[_attr].optional or (_attr in values and values[_attr]):
                _value = _type(values[_attr]) if _type else values[_attr] 
            setattr(obj, _attr, _value)
    
    def add_request(self, obj):
        # Calculate only if it is not present
        if not hasattr(obj, 'requested_nodes'):
            _partition = 0 
            for _res in self.system_resources:
                _total_request = getattr(obj, _res)
                assert(_total_request >= 0), 'The request for {} is no feasible ({}). Accepted values are equal or greater than 0. Job {} must be tweaked before re-run. See the example.'.format(_res, _total_request, obj.id)
                _partition = max([_partition] + [round(getattr(obj, _res) / _resources[_res]) for _resources in self.group_resources.values()])
            setattr(obj, 'requested_nodes', _partition)
        if not hasattr(obj, 'requested_resources'):
            _partition = getattr(obj, 'requested_nodes')
            setattr(obj, 'requested_resources', {_res: getattr(obj, _res) // _partition for _res in self.system_resources})
    
    def default_job_description(self):
        # Attribute to identify the user
        user_id = attribute_type('user_id', int)
        
        # New attributes required by the Dispatching methods.
        expected_duration = attribute_type('expected_duration', int)
        
        # Default system resources: core and mem.
        total_cores = attribute_type('core', int)
        total_mem = attribute_type('mem', int)
    
        # This attributes are required to be set, if not by default are calculated. As in this example (explained in tweak_dict function) 
        requested_nodes = attribute_type('requested_nodes', int)
        requested_resources = attribute_type('requested_resources', dict)
        
        return [total_cores, total_mem, requested_nodes, requested_resources, expected_duration, user_id ]
            
class event_mapper:
    """
        This class coordinates events submission, queueing and ending. 
        kwargs:
            - nothing by the moment
    """
    def __init__(self, _resource_manager, **kwargs):
        assert(isinstance(_resource_manager, resource_manager)), 'Wrong type for the resource_manager argument.'
        self.resource_manager = _resource_manager
        self.constants = CONSTANT()
        #=======================================================================
        # if 'resource_manager' in kwargs:            
        #     self.resource_manager = kwargs['resource_manager']
        #=======================================================================
        # Stats
        self.first_time_dispatch = None
        self.last_run_time = None
        self.slowdowns = []
        self.wtimes = []
        
        self.current_time = 0
        self.time_points = sorted_list()
        # self.ending_time_points = sorted_list()
        self.events = {}
        self.loaded = {}
        self.queued = []
        self.real_ending = {}
        self.running = []
        self.finished = []
            
    def load_events(self, es):
        if isinstance(es, list):
            for e in es:
                self.load_event(e)
        else:
            self.load_event(es)
                    
    def load_event(self, e):
        assert(isinstance(e, event)), 'Using %s, expecting a single %s' % (e.__class__, event.__name__)
        if self.current_time == e.queued_time:
            self.submit_event(e)
        elif self.current_time < e.queued_time:
            if e.queued_time not in self.loaded:
                 self.loaded[e.queued_time] = []
            self.loaded[e.queued_time].append(e.id)
            self.time_points.add(e.queued_time)
        else:
            raise Exception('Time sync problem, the actual event was loaded after the real submit time. This a programming error, must be checked.')
                        

    def move_to_finished(self, events_dict):
        """
            There are two time points for a job could ends, the expected one and the real one. 
            The job must run until the real one is reached, then if a job is waiting to finish but is less than the
            real ending time, this value must be updated with the real one.        
        """
        _es = []
        for e_id in self.real_ending.pop(self.current_time, []):
            if e_id in self.running:
                self.running.remove(e_id)
                e = events_dict[e_id]
                self.finish_event(e)  
                _es.append(e_id)
        self.last_run_time = self.current_time
        return _es

    def finish_event(self, e):
        e.end_time = self.current_time
        e.running_time = e.end_time - e.start_time
        e.waiting_time = e.start_time - e.queued_time
        e.slowdown = float("{0:.2f}".format((e.waiting_time + e.running_time) / e.running_time)) if e.running_time != 0 else e.waiting_time if e.waiting_time != 0 else 1.0
        self.slowdowns.append(e.slowdown)   
        self.wtimes.append(e.waiting_time)
        self.finished.append(e.id)
        e.end_order = len(self.finished)
        asyncio.create_subprocess_exec([e.schd_write_out(), e.schd_pprint_write_out()])
   
    def dispatch_event(self, _e, time, time_diff, _nodes): 
        id = _e.id
        start_time = time + time_diff
        assert(self.current_time == start_time), 'Start time is different to the current time'
        
        """
         Update job info
        """
        _e.start_time = start_time
        _e.assigned_nodes = _nodes


        # Used only for statistics
        if self.first_time_dispatch == None:
            self.first_time_dispatch = start_time

        if _e.duration == 0:
            logging.debug('%s: %s Dispatched and Finished at the same moment. Job Lenght 0' % (self.current_time, id))
            self.finish_event(_e)
            self.time_points.add(self.current_time)
            return False
        # Move to running jobs 
        self.running.append(id)
        """
            Setting the ending time as walltime
        """
        expected_end_time = _e.start_time + _e.expected_duration
        real_end_time = _e.start_time + _e.duration
        # self.ending_time_points.add(expected_end_time, real_end_time)
        self.time_points.add(expected_end_time)
        self.time_points.add(real_end_time)

        if real_end_time not in self.real_ending:
            self.real_ending[real_end_time] = [] 
        self.real_ending[real_end_time].append(id)
        logging.debug('%s: %s Dispatched! Init %s Ending %s' % (self.current_time, id, start_time, expected_end_time))
        return True
        #=======================================================================
        # elif self.current_time <= start_time:
        #     # Here must be reserved slots!!
        #     logging.debug('%s: %s moved to Queue! Will init %s' % (self.current_time, id, start_time))
        #     self.submit_event(_e, start_time)
        #     return False
        # else:
        #     logging.error('Error: ', self.current_time, time)
        #     raise Exception('Can\'t comeback to the past :( (%s->%s)' % (self.current_time, time))
        #=======================================================================
        
    def submit_event(self, e_id):  # (self, e, re_queued_time=None, debug=False):
        # assert(isinstance(e, event)) 
        # self.queued.append(e.id)
        self.queued.append(e_id)
    
    def next_events(self):
        self.current_time = self.time_points.pop()
        if self.current_time is None:
            return []
        submitted = self.loaded.pop(self.current_time, [])
        new_queue = self.queued + submitted 
        logging.debug('Next events: \n-Recently submited: %s\n-Already queued: %s' % (submitted, self.queued))
        self.queued.clear()
        return new_queue
        
    def has_events(self):
        return (self.loaded or self.queued or self.running)
    
    def show_map(self, events):
        i = 1
        print('#\tjob_id\t\twtime\trtime\tbsld\tnodes\tcore\tgpu\tmic\tmem\tnodes')
        self.finished.sort(key=lambda e: events[e].start_time, reverse=False)
        for e in self.finished:
            _e = events[e]
            _e_resources = _e.requested_resources
            _wtime = _e.start_time - _e.queued_time
            _rtime = _e.end_time - _e.start_time
            _bsld = (_wtime + _rtime) / _rtime if _rtime > 0 else 1.0
            print('%s\t%s\t%s\t%s\t%.2f\t%s\t%s\t%s\t%s\t%s\t%s' % (i, _e.id, _wtime, _rtime, _bsld, _e.requested_nodes, _e_resources['core'], _e_resources['gpu'], _e_resources['mic'], _e_resources['mem'], ','.join([n.split('_')[1] for n in _e.assigned_nodes])))
            i += 1        
 
    def dispatch_events(self, event_dict, to_dispatch, time_diff, _debug=False):
        for (_time, _id, _nodes) in to_dispatch:
            assert(_time is None or _time >= self.current_time), 'Receiving wrong schedules.'
            """
            Time must be equal or later than current time.
                Equals will be dispatched in the momement, instead later ones, which will be requeued with expected ending time of the job that release the resources. 
            If the expected ending is surpass, because the job takes more time to finish, the time tuple\'s element must be None. '
            """
            if not _nodes:
                # For blocked jobs
                if _time is not None and _time != self.current_time:
                    self.time_points.add(_time)
                """
                 Maintaining the event in the queue 
                """
                self.submit_event(_id)
                continue
            _e = event_dict[_id]
            # print(_e, _time, time_diff, _nodes)
            if self.dispatch_event(_e, _time, time_diff, _nodes):
                self.resource_manager.allocate_event(_e, _nodes)
                
    def release_ended_events(self, event_dict):
        _es = self.move_to_finished(event_dict)
        for _e in _es:
            self.resource_manager.remove_event(_e)
            # Freeing mem (finished events)
            event_dict.pop(_e)
        return _es
    
    def simulated_status(self):
        return ('Loaded {}, Queued {}, Running {}, and Finished {} Jobs'.format(len(self.loaded), len(self.queued), len(self.running), len(self.finished)))
    
    def availability(self):
        return self.resource_manager.availability()
    
    def usage(self):
        return self.resource_manager.resources.usage()
    
    def simulated_current_time(self):
        return self.current_time
    
    def __str__(self):
        return 'Loaded: %s\nQueued: %s\nRunning: %s\nExpected job finish: %s\nReal job finish on: %s,\nFinished: %s\nNext time events: %s' % (self.loaded, self.queued, self.running, None, self.real_ending, self.finished, self.time_points)



"""
MIT License

Copyright (c) 2017 cgalleguillosm, AlessioNetti

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import re
import os
import logging
import random
import json
from datetime import datetime
import time
from threading import Timer, Thread
from inspect import getouterframes, currentframe
import functools
import inspect
from collections import namedtuple, Mapping
from _collections import deque
from math import sqrt
from bisect import bisect, bisect_left, bisect_right
import socket
from sys import platform as _platform, exit as _exit
import threading
from abc import ABC, abstractmethod
from _functools import reduce
from itertools import islice
from builtins import int, str
import ntpath
import sys


_swf_int_pattern = ('\s*(?P<{}>[-+]?\d+)', int)
_swf_float_pattern = ('\s*(?P<{}>[-+]?\d+\.\d+|[-+]?\d+)', float)
_swf_avoid_regexps = [r'^;.*']
default_swf_parse_config = (
    {
        'job_number': _swf_int_pattern,
        'submit_time': _swf_int_pattern,
        'wait_time': _swf_int_pattern,
        'duration': _swf_int_pattern,
        'allocated_processors': _swf_int_pattern,
        'avg_cpu_time': _swf_float_pattern,
        'used_memory': _swf_int_pattern,
        'requested_number_processors': _swf_int_pattern,
        'requested_time': _swf_int_pattern,
        'requested_memory': _swf_int_pattern,
        'status': _swf_int_pattern,
        'user_id': _swf_int_pattern,
        'group_id': _swf_int_pattern,
        'executable_number': _swf_int_pattern,
        'queue_number': _swf_int_pattern,
        'partition_number': _swf_int_pattern,
        'preceding_job_number': _swf_int_pattern,
        'think_time_prejob': _swf_int_pattern
    }, _swf_avoid_regexps)

default_swf_mapper = {
    'job_number': 'job_id',
    'submit_time': 'queued_time',
    'requested_time': 'expected_duration'
}

def default_sorting_function(obj1, obj2, avoid_data_tokens=[';']): 
    if obj1[0] in avoid_data_tokens or obj2[0] in avoid_data_tokens:
        return 1
    return default_sorted_attribute(obj1) - default_sorted_attribute(obj2) 

def default_sorted_attribute(workload_line, attr='submit_time', converter=None):
    # print('wl: ', workload_line)
    value = workload_parser(workload_line, attr)[attr]
    if converter:
        return converter(value)
    return value

def workload_parser(workload_line, attrs=None, avoid_data_tokens=[';']):
    """ 
        Attributes of each workload line in a SWF format (separated by space):
        
        1. job_number -- a counter field, starting from 1.
        2. submit_time -- in seconds. The earliest time the log refers to is zero, and is usually the submittal time of the first job. The lines in the log are sorted by ascending submittal times. It makes sense for jobs to also be numbered in this order.
        3. wait_time -- in seconds. The difference between the job's submit time and the time at which it actually began to run. Naturally, this is only relevant to real logs, not to models.
        4. duration -- in seconds. The wall clock time the job was running (end time minus start time).
        5. allocated_processors -- an integer. In most cases this is also the number of processors the job uses; if the job does not use all of them, we typically don't know about it.
        6. avg_cpu_time -- Time Used for both user and system, in seconds. This is the average over all processors of the CPU time used, and may therefore be smaller than the wall clock runtime. If a log contains the total CPU time used by all the processors, it is divided by the number of allocated processors to derive the average.
        7. used_memory -- in kilobytes. This is again the average per processor.
        8. requested_number_processors --- Requested Number of Processors.
        9. requested_time -- This can be either runtime (measured in wallclock seconds), or average CPU time per processor (also in seconds) -- the exact meaning is determined by a header comment. In many logs this field is used for the user runtime estimate (or upper bound) used in backfilling. If a log contains a request for total CPU time, it is divided by the number of requested processors.
        10. requested_memory -- Requested memory in kilobytes per processor.
        11. status -- 1 if the job was completed, 0 if it failed, and 5 if cancelled. If information about chekcpointing or swapping is included, other values are also possible. See usage note below. This field is meaningless for models, so would be -1.
        12. user_id -- a natural number, between one and the number of different users.
        13. group_id -- a natural number, between one and the number of different groups. Some systems control resource usage by groups rather than by individual users.
        14. executable_number -- a natural number, between one and the number of different applications appearing in the workload. in some logs, this might represent a script file used to run jobs rather than the executable directly; this should be noted in a header comment.
        15. queue_number -- a natural number, between one and the number of different queues in the system. The nature of the system's queues should be explained in a header comment. This field is where batch and interactive jobs should be differentiated: we suggest the convention of denoting interactive jobs by 0.
        16. partition_number -- a natural number, between one and the number of different partitions in the systems. The nature of the system's partitions should be explained in a header comment. For example, it is possible to use partition numbers to identify which machine in a cluster was used.
        17. preceding_job_number -- this is the number of a previous job in the workload, such that the current job can only start after the termination of this preceding job. Together with the next field, this allows the workload to include feedback as described below.
        18. think_time_prejob -- this is the number of seconds that should elapse between the termination of the preceding job and the submittal of this one.
    """ 
    if workload_line[0] in avoid_data_tokens:
        return workload_line
    _common_int_pattern = ('\s*(?P<{}>[-+]?\d+)', int)
    _common_float_pattern = ('\s*(?P<{}>[-+]?\d+\.\d+|[-+]?\d+)', float)
    _dict = {
        'job_number': _common_int_pattern,
        'submit_time': _common_int_pattern,
        'wait_time': _common_int_pattern,
        'duration': _common_int_pattern,
        'allocated_processors': _common_int_pattern,
        'avg_cpu_time': _common_float_pattern,
        'used_memory': _common_int_pattern,
        'requested_number_processors': _common_int_pattern,
        'requested_time': _common_int_pattern,
        'requested_memory': _common_int_pattern,
        'status': _common_int_pattern,
        'user_id': _common_int_pattern,
        'group_id': _common_int_pattern,
        'executable_number': _common_int_pattern,
        'queue_number': _common_int_pattern,
        'partition_number': _common_int_pattern,
        'preceding_job_number': _common_int_pattern,
        'think_time_prejob': _common_int_pattern
    }
    _sequence = _dict.keys() if not attrs else ((attrs,) if isinstance(attrs, str) else attrs)
    reg_exp = r''
    for _key in _sequence:
        reg_exp += _dict[_key][0].format(_key)
    p = re.compile(reg_exp)
    _matches = p.match(workload_line)
    # value_func = lambda line, _dict, key: 
    _dict_line = _matches.groupdict()
    return {key:_dict[key][1](_dict_line[key]) for key in _sequence} 
    

def sort_file(input_filepath, lines=None, sort_function=default_sorting_function, avoid_data_tokens=[';'], output_filepath=None):
    """
        The input file for the simulator must be sorted by submit time. It modifies the file input file, 
        or also can be saved to a new one if the output_filepath arg is defined.
          
        :param input_filepath: Input workload file
        :param lines: Number of lines to be read. It includes all lines from the begining of the file. 
        :param sort_function: (Optional) The function that sorts the file by submit time. The user is responsable 
                to define the correct function. If a workload with SWF format is used, by default 
                default_sorting_function (SWF workload) is used.
        :param avoid_data_tokens: (Optional) By default avoid to modify comment lines of SWF workload.      
        :param output_filepath: (Optional) The sorted data is saves into another file (this filepath). 
                It will not content the lines that begin with tokens of the avoid_data_tokens var.
        
        :return: A list of queued time points.  
    """
    assert(callable(sort_function))
    logging.debug('Sorting File: %s ' % (input_filepath))
    with open(input_filepath) as f:
        sorted_file = list(f if not lines else islice(f, lines))
        sorted_file .sort(
            key=cmp_to_key(sort_function)
        )
    if output_filepath is None:
        output_filepath = input_filepath
    logging.debug("Writing sorted file to %s" % (output_filepath))
    queued_times = sorted_list()
    with open(output_filepath, 'w') as f:
        for line in sorted_file:
            if line[0] in avoid_data_tokens:
                f.write(line)
                continue
            _line = workload_parser(line)
            if int(_line['requested_number_processors']) == -1 and int(_line['allocated_processors']) == -1 or int(_line['requested_memory']) == -1 and int(_line['used_memory']) == -1:
                continue   
            qtime = default_sorted_attribute(line, 'submit_time')
            queued_times.add(qtime)
            f.write(line)
    return queued_times.get_list()  
    
def cmp_to_key(mycmp):
    'Convert a cmp= function into a key= function'
    class k(object):
        def __init__(self, obj, *args):
            self.obj = obj
        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0
        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0
        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0
        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0  
        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0
        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0
    return k    

def from_isodatetime_2_timestamp(timestamp):
        p = re.compile(r'(\d{4})-(\d{2})-(\d{2})\s(\d{2}):(\d{2}):(\d{2})')
        m = p.search(timestamp).groups()
        # year, month, day, hour, minute, second, microsecond
        t = datetime(year=int(m[0]), month=int(m[1]), day=int(m[2]), hour=int(m[3]), minute=int(m[4]), second=int(m[5]))
        return int(t.timestamp())

def sorted_attr(line, reg_exp=(r'\d+\.\w+;.+;\w+@\w+\.eurora\.cineca\.it;\w+;(?P<queue_time>\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2})__'), group_name='queue_time'):
    p = re.compile(reg_exp)
    return p.match(line).groupdict()[group_name]

def sort_job_function(x, y):
    return from_isodatetime_2_timestamp(sorted_attr(x)) - from_isodatetime_2_timestamp(sorted_attr(y))      

class time_demon:
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self._timer = Timer(self.interval, self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False
        
class watcher_demon:

    MAX_LENGTH = 2048
    PRINT_INTERVAL = 300
    
    def __init__(self, port, functions):
        self.server_address = ('', port)
        af = socket.AF_INET
        self.sock = socket.socket(af, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.settimeout(1)
        self.thread = None
        self.timedemon = None
        self.hastofinish = False
        self.const = CONSTANT()
        self.functions = functions


    def start(self):
        self.thread = threading.Thread(target=self.listenForRequests)
        self.hastofinish = False
        self.timedemon = time_demon(self.PRINT_INTERVAL, self.resourceusageprint)
        self.thread.start()
        self.timedemon.start()

    def listenForRequests(self):
        # Listen for incoming connections
        # Reusing
        self.sock.bind(self.server_address)
        self.sock.listen(5)

        while not self.hastofinish:
            try:
                connection, client_address = self.sock.accept()
                with connection:
                    # print('connection from %s' % (client_address[0]))
                    data = json.loads(connection.recv(self.MAX_LENGTH).decode())
                    if isinstance(data, str):
                        response = {}
                        response['actual_time'] = str(str_datetime(self.call_inner_function('current_time_function')))
                        if data == 'progress':
                            response['input_filepath'] = self.const.input_filepath
                            response['progress'] = os.path.getsize(self.const.sched_output_filepath) / os.path.getsize(self.const.input_filepath)
                            response['time'] = time.clock() - self.const.start_time
                        elif data == 'usage':
                            response['simulation_status'] = self.call_inner_function('simulated_status_function')
                            response['usage'] = self.call_inner_function('usage_function')
                        elif data == 'all':
                            response['input_filepath'] = self.const.input_filepath
                            response['progress'] = os.path.getsize(self.const.sched_output_filepath) / os.path.getsize(self.const.input_filepath)
                            response['time'] = time.clock() - self.const.start_time
                            response['simulation_status'] = self.call_inner_function('simulated_status_function')
                            response['usage'] = self.call_inner_function('usage_function')
                        connection.sendall(json.dumps(response).encode())
                    connection.close()
            except socket.timeout:
                pass
        self.sock.close()
        
    def call_inner_function(self, name):
        if name in self.functions:
            _func = self.functions[name]
            if callable(_func):
                return _func()
            else:
                return _func
        raise Exception('{} was no defined'.format(name)) 

    def stop(self):
        self.hastofinish = True
        self.timedemon.stop()

    def resourceusageprint(self):
        rm = self.const.resource_manager_instance
        if rm is not None and rm.resources is not None:
            print('- ' + rm.resources.usage())


def generate_config(config_fp, **kwargs):
    _local = {}
    for k, v in kwargs.items():
        print(k, v)
        _local[k] = v
    with open(config_fp, 'w') as c:
        json.dump(_local, c, indent=2)

def hinted_tuple_hook(obj):
    if '__tuple__' in obj:
        return tuple(obj['items'])
    else:
        return obj

def load_config(config_fp):
    _dict = None
    with open(config_fp) as c:
        _dict = json.load(c, object_hook=hinted_tuple_hook)
    return _dict

def simulate_event(min_time, max_time):
    time.sleep(random.randint(min_time, max_time))
    
class Singleton(object):
    _instances = {}

    def __new__(class_, *args, **kwargs):
        if class_ not in class_._instances:
            class_._instances[class_] = super(Singleton, class_).__new__(class_, *args, **kwargs)
        return class_._instances[class_]

class CONSTANT(Singleton):
    """
        This class allows to load all config into the Singleton Object called CONSTANT. 
        For accessing to all the parameters, it will be possible only calling the parameters as its attribute.
    
        I.e:
        Config:
        PATH = '/path/to/'
        
        Program
        c = CONSTANT()
        print(c.PATH)
        
        >> /path/to/
    
        It's loaded into all base class by default!
        New attrs could be passed as dict (load_constants) or simply with (attr, value) (load_constant)
    """    
    def load_constants(self, _dict):
        for k, v in _dict.items():
            self.load_constant(k, v)
            
    def load_constant(self, k, v):
        assert(not hasattr(self, k)), '{} already exists as constant ({}={}). Choose a new name.'.format(k, k, getattr(self, k))
        setattr(self, k, v)

    #===========================================================================
    # An idea for improving the config is use the class name of parameters, ths could be controlled with the following method, but how??  
    # def __getattribute__(self, *args, **kwargs):
    #     return object.__getattribute__(self, *args, **kwargs)
    #===========================================================================


class str_:
    def __init__(self, text):
        self.text = text
    
    def __str__(self):
        return self.text


class str_datetime:
    
    def __init__(self, epoch_time):
        self.str_datetime = datetime.fromtimestamp(int(epoch_time)).strftime('%Y-%m-%d %H:%M:%S')
        
    def __format__(self, *args):
        return self.str_datetime
    
    def __str__(self):
        return self.str_datetime


class str_time:
    
    def __init__(self, secs):
        self.str_time = time.gmtime(int(secs))  # time.strftime('%H:%M:%S', time.gmtime(int(secs)))
        
    def __str__(self):
        return self.str_time
    
class str_resources:
    
    def __init__(self, nodes, resources):
        self.nodes = nodes
        self.resources = resources  # namedtuple('resources', [k for k in resources.keys()])(**resources)
        self.constants = CONSTANT()
        if hasattr(self.constants, 'resource_order'):
            self.order = getattr(self.constants, 'resource_order')
        else:
            self.order = list (self.resources.keys())
        
    def __str__(self):
        #=======================================================================
        # r = self.resources
        # print(self.nodes, r)
        # return '#'.join([';'.join([node.split('_')[1], str(r.core), str(r.gpu), str(r.mic), str(r.mem)]) for node in self.nodes]) + '#'
        #=======================================================================
        return '#'.join([';'.join([node.split('_')[1]] + [str(self.resources[_k]) for _k in self.order]) for node in self.nodes]) + '#'

class str_nodes:
    
    def __init__(self, nodes):
        self.nodes = nodes
    
    def __format__(self, format_spec):
        return self.__str__()
    
    def __str__(self):
        return ','.join([node.split('_')[1] for node in self.nodes])



class sorted_object_list():
    def __init__(self, sorting_priority, class_obj=None, _list=[]):
        assert(isinstance(sorting_priority, dict) and set(['main', 'break_tie']) <= set(sorting_priority.keys()))

        self.main_sort = sorting_priority['main']
        self.break_tie_sort = sorting_priority['break_tie']
        self.list = []
        self.main = []
        self.secondary = []
        self.map = {
            'pos': {},
            'id': {}
        }
        self.objects = {}
        # dict values, function or inner attributes of wrappred objs
        self._iter_func = lambda act, next: act.get(next) if isinstance(act, dict) else (getattr(act, next)() if callable(getattr(act, next)) else getattr(act, next))

        if _list:
            self.add(*_list)
            
    # @abstractmethod
    def add(self, *args):
        for arg in args:
            _id = getattr(arg, 'id')
            if _id in self.map['id']:
                continue
            self.objects[_id] = arg
            _main = reduce(self._iter_func, self.main_sort.split('.'), arg)  # getattr(arg, self.main_sort)
            _sec = reduce(self._iter_func, self.break_tie_sort.split('.'), arg)
            _pos = bisect_left(self.main, _main)
            main_pos_r = bisect_right(self.main, _main)
            if _pos == main_pos_r:
                self.list.insert(_pos, _id)
                self.main.insert(_pos, _main)
                self.secondary.insert(_pos, _sec)
            else:
                _pos = bisect_left(self.secondary[_pos :main_pos_r], _sec) + _pos
                self.list.insert(_pos, _id)
                self.main.insert(_pos, _main)
                self.secondary.insert(_pos, _sec)
            self.map_insert(self.map['id'], self.map['pos'], _pos, _id)
    
    def _get_value(self, arg, attr):
        
        if isinstance(self.break_tie_sort, tuple):
            _sec = arg
            for _attr in self.break_tie_sort:
                _sec = getattr(_sec, _attr)
        else:
            _sec = getattr(arg, self.break_tie_sort)
    
    def map_insert(self, ids_, poss_, new_pos, new_id):
        n_items = len(ids_)
        if n_items > 0:
            if not(new_pos in poss_):
                poss_[new_pos] = new_id
                ids_[new_id] = new_pos
            else:
                self.make_map(ids_, poss_, new_pos)
        else:
            ids_[new_id] = new_pos
            poss_[new_pos] = new_id
    
    # Improve this
    def make_map(self, ids_, poss_, new_pos=0, debug=False):    
        for _idx, _id in enumerate(self.list[new_pos:]):
            ids_[_id] = _idx + new_pos
            poss_[_idx + new_pos] = _id
        if len(ids_) == len(poss_):
            return
        for p in list(poss_.keys()):
            if p > _idx:
                del poss_[p] 
        
    def remove(self, *args, **kwargs):
        for id in args:
            assert(id in self.objects)
            del self.objects[id]            
            self._remove(self.map['id'][id], **kwargs)
            print(self.map)
            
    def _remove(self, _pos, **kwargs):
        del self.list[_pos]
        del self.secondary[_pos]
        del self.main[_pos]
        
        _id = self.map['pos'].pop(_pos)
        del self.map['id'][_id]
        self.make_map(self.map['id'], self.map['pos'], **kwargs)

                
    def get(self, pos):
        return self.list[pos]

    def get_object(self, id):
        return self.objects[id]
     
    def get_list(self):
        return self.list
    
    def get_object_list(self):
        return [self.objects[_id] for _id in self.list] 
                
    def __len__(self):
        return len(self.list)
    
    # Return None if there is no coincidence
    def pop(self, id=None, pos=None):
        assert(not all([id, pos])), 'Pop only accepts one or zero arguments'
        if not self.list:
            return None
        elif id:
            return self._specific_pop_id(id)
        elif pos:
            return self._specific_pop_pos(pos)
        else:
            _id = self.list[0]
            self._remove(0)
            return self.objects.pop(_id)        
    
    def _specific_pop_id(self, id):
        _obj = self.objects.pop(id, None)
        if _obj:
            self._remove(self.map['id'][id])
        return _obj
    
    def _specific_pop_pos(self, pos):
        _id = self.map['pos'].pop(pos, None)
        if _id:
            self.map['pos'][pos] = _id
            self._remove(pos)
        return self.objects.pop(_id, None)        

    def __iter__(self):
        self.actual_index = 0
        return self

    def __next__(self):
        try:
            self.actual_index += 1
            return self.list[self.actual_index - 1]
        except IndexError:
            raise StopIteration
    
    def get_reversed_list(self):
        return list(reversed(self.list))
    
    def get_reversed_object_list(self):
        return [ self.objects[_id] for _id in reversed(self.list)]
    
    def __str__(self):
        return str(self.list)
    
class sorted_list:
    def __init__(self, _list=[]):
        assert(isinstance(_list, (list)))
        self.list = []
        if _list:
            self.add(*_list)
    
    def add(self, *args):
        for arg in args:
            if len(self.list) == 0:
                self.list.append(arg)
            else:
                _num = arg
                pos = bisect(self.list, _num)
                #===============================================================
                # mid = int(len(self.list) // 2)
                # if self.list[mid] > _num:
                #     pos = bisect(self.list, _num, hi=mid)
                # else:
                #     pos = bisect(self.list, _num, hi=mid)
                #===============================================================
                if self.list[pos - 1] != _num:
                    self.list.insert(pos, _num)
                    
    def get_list(self):
        return self.list
   
    def find(self, _num):
        return bisect(self.list, _num) - 1
    
    def remove(self, *args):
        for arg in args:
            self.list.remove(arg)
                
    def get(self, pos):
        return self.list[pos]
                
    def __len__(self):
        return len(self.list)
    
    def pop(self):
        if self.list:
            return self.list.pop(0)
        return None

    def __iter__(self):
        self.actual_index = 0
        return self

    def __next__(self):
        try:
            self.actual_index += 1
            return self.list[self.actual_index - 1]
        except IndexError:
            raise StopIteration
    
    def __str__(self):
        return str(self.list)
    
    def _check_sort(self):
        
        for i in range(len(self.list) - 1):
            if self.list[i] >= self.list[i + 1]:
                self.list[0:i + 1]
                raise Exception('Sorting problem!')
            
class FrozenDict(Mapping):
    """
        Inmutable dictionary useful for storing parameter that are dinamycally loaded
    """
    def __init__(self, *args, **kwargs):
        self._d = dict(*args, **kwargs)
        self._hash = None

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getitem__(self, key):
        return self._d[key]

    def __hash__(self):
        if self._hash is None:
            self._hash = 0
            for pair in self.iteritems():
                self._hash ^= hash(pair)
        return self._hash
    
def clean_results(*args):
    for fp in args:
        if os.path.isfile(fp) and os.path.exists(fp):
            os.remove(fp)
            
class DEFAULT_SIMULATION:
    
    parameters = {
        "CONFIG_FOLDER_NAME": "config/",
        "RESULTS_FOLDER_NAME": "results/",
        "SCHEDULE_OUTPUT": {
            "format": "{job_id};{user};{queue_time}__{assignations}__{start_time};{end_time};{total_nodes};{total_cpu};{total_mem};{expected_duration};",
            "attributes": {
                "job_id": ("id", "str"),
                "user": ("user_id", "str"),
                "queue_time": ("queued_time", "accasim.utils.misc.str_datetime"),
                "start_time": ("start_time", "accasim.utils.misc.str_datetime"),
                "end_time": ("end_time", "accasim.utils.misc.str_datetime"),
                "assignations": ("assigned_nodes", "requested_resources", "accasim.utils.misc.str_resources"),
                "total_nodes": ("requested_nodes", "int"),
                "total_cpu": ("core", "int"),
                "total_mem": ("mem", "int"),
                "expected_duration": ("expected_duration", "int")      
            }
        },
        "PPRINT_SCHEDULE_OUTPUT": {
            "format": "{:>5} {:>15} {:^19} {:^19} {:>8} {:>8} {:>8} {:>5} {:>4} {:>10} {:<20}",
            "order": ["n", "job_id", "start_time", "end_time", "wtime", "rtime", "slowdown", "nodes", "core", "mem", "assigned_nodes"],
            "attributes":{
                "n": ("end_order", "int"),
                "job_id": ("id", "str"),
                "start_time": ("start_time", "accasim.utils.misc.str_datetime"),
                "end_time": ("end_time", "accasim.utils.misc.str_datetime"),
                "wtime": ("waiting_time", "int"),
                "rtime": ("running_time", "int"),
                "slowdown": ("slowdown", "float"),
                "nodes": ("requested_nodes", "int"),
                "core": ("core", "int"),
                "mem": ("mem", "int"),
                "assigned_nodes": ("assigned_nodes", "accasim.utils.misc.str_nodes")
            }
        },
        "SCHED_PREFIX": "sched-",
        "PPRINT_PREFIX": "pprint-",
        "STATISTICS_PREFIX": "stats-",
        "BENCHMARK_PREFIX": "bench-",
        "RESOURCE_ORDER": ["core", "mem"],
        "WATCH_PORT": 8999
    }
    
def path_leaf(path):
    head, tail = ntpath.split(path)
    return (head, tail or ntpath.basename(head))
            
if __name__ == '__main__2':
    Resource = namedtuple('Resource', ['q', 'w'])
    Job = namedtuple('Job', ['id', 'resources', 'vars'])
    lista = [
        Job(**{'id': 'job.1', 'resources': Resource(**{'q': '5', 'w':'2'}), 'vars': {'w': 2} }),
        Job(**{'id': 'job.2', 'resources': Resource(**{'q': '5', 'w':'1'}), 'vars': {'w': 1} }),
        Job(**{'id': 'job.3', 'resources': Resource(**{'q': '3', 'w':'1'}), 'vars': {'w': 1} }),
        Job(**{'id': 'job.4', 'resources': Resource(**{'q': '1', 'w':'10'}), 'vars': {'w': 10} }),
        Job(**{'id': 'job.5', 'resources': Resource(**{'q': '1', 'w':'1'}), 'vars': {'w': 1} }),
        Job(**{'id': 'job.6', 'resources': Resource(**{'q': '6', 'w':'1'}), 'vars': {'w': 1} }),
        Job(**{'id': 'job.7', 'resources': Resource(**{'q': '6', 'w':'7'}), 'vars': {'w': 7} }),
        Job(**{'id': 'job.8', 'resources': Resource(**{'q': '4', 'w':'2'}), 'vars': {'w': 2} }),
        Job(**{'id': 'job.9', 'resources': Resource(**{'q': '4', 'w':'1'}), 'vars': {'w': 1} }),
        ]
    sorting_priority = {'main': 'resources.q', 'break_tie': 'vars.w'}
    s_obj = Job
    slist = sorted_object_list(sorting_priority, s_obj)
    slist.add(*lista)
    print('All elements: ', slist)
    print('Inverse List: ', slist.get_reversed_list())
    _pop = slist.pop()
    print('Pop : ', _pop, '. Remaining list: ', slist)
    _pop = slist.pop(pos=1)
    print('Pop : ', _pop, '. Remaining list: ', slist)
    _pop = slist.pop(id='job.2')
    print('Pop : ', _pop, '. Remaining list: ', slist)
    _id = 'job.1'
    slist.remove(_id, debug=True)
    print('Removing : ', _id, '. Remaining list: ', slist)
    while slist.pop():
        print(slist)
            
if __name__ == '__main__2':
    Job = namedtuple('Job', ['id', 'q', 'w'])
    lista = [
        Job(**{'id': 'job.1', 'q': '5', 'w':'2'}),
        Job(**{'id': 'job.2', 'q': '5', 'w':'1'}),
        Job(**{'id': 'job.3', 'q': '3', 'w':'1'}),
        Job(**{'id': 'job.4', 'q': '1', 'w':'10'}),
        Job(**{'id': 'job.5', 'q': '1', 'w':'1'}),
        Job(**{'id': 'job.6', 'q': '6', 'w':'1'}),
        Job(**{'id': 'job.7', 'q': '6', 'w':'7'}),
        Job(**{'id': 'job.8', 'q': '4', 'w':'2'}),
        Job(**{'id': 'job.9', 'q': '4', 'w':'1'}),
        ]
    sorting_priority = {'main': 'q', 'break_tie':'w'}
    s_obj = Job
    slist = sorted_object_list(sorting_priority, s_obj)
    slist.add(*lista)
    print(slist)
    print(slist.reverse())
    if 'job.8' in slist:
        print('si')
    slist.remove(*['job.8'])
    print(slist)
    if 'job.8' not in slist:
        print('no')
    
if __name__ == '__main__':
    random.seed(0)
    lista = sorted_list([])
    
    max_val = 10000
    values = random.sample(range(1, max_val * 2), max_val)
    random.shuffle(values)
    print('Adding new elements and resorting everytime')
    i_time = time.clock()
    for _v in values:
        lista.add(_v)
    print('Custom sorting time: ', time.clock() - i_time, len(lista.get_list()))
    
    i_time = time.clock()
    inc_list = []
    for i in values:
        inc_list.append(i)
        sorted(inc_list)
    print('Sorting time: ', time.clock() - i_time, len(values))
    
    print('Full list sorting')
    lista = sorted_list([])
    i_time = time.clock()
    lista.add(*values)
    print('Custom sorting time: ', time.clock() - i_time, len(lista.get_list()))
    
    i_time = time.clock()
    sorted(values)
    print('Sorting time: ', time.clock() - i_time, len(values))



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.cm as mplcm
import matplotlib.colors as colors
import time
import matplotlib.animation as Animation
from accasim.utils.misc import sorted_object_list, str_datetime
from collections import namedtuple
from threading import Thread

class scheduling_visualization:
    
    def __init__(self, datasource, system_info, min_time=10 * 60, max_time=30 * 60):
        self.datasource = datasource
        self.system_info = system_info
        self.min_time = min_time
        self.max_time = max_time
        self.job_patches = {}
        self.job_object = {}
        self.job_resouce_type = {}
        self.sorted_jobs = {}
        self.job_color = {}
        for key in self.system_info.keys():
            self.job_resouce_type[key] = namedtuple('Resource_{}'.format(key), ['id', key, 'expected_duration'])
            self.sorted_jobs[key] = sorted_object_list({'main': key, 'break_tie':'expected_duration'}, None)
        self.figure = None
        self.animation = None
        self.axis_mapper = None
        self.color_pos = None
        self.last_line = {'time': None}
        self.color_blacklist = ['red']
        self.canvas = None
        
        # Create initial structure for ploting
        self.set_colors()
        self.running = False

    def start(self):
        self.running = True
        self.create_empty_map()

    def stop(self):
        if self.running:
            self.running = False
            try:
                self.animation._stop()
                self.animation.repeat = False
            except AttributeError as e:
                print(e)
        plt.close()        
        
    def load_new_jobs(self, current_time, running_jobs, _resources):
        for _id, _job in running_jobs.items():
            if _id in self.job_patches:
                continue
            self.job_object[_id] = _job
            for res in _resources:
                # self.sorted_jobs[res].add(self.job_resouce_type[res](_id, _job.get_total_requested_resource()[res], _job.expected_duration))
                self.sorted_jobs[res].add(self.job_resouce_type[res](_id, getattr(_job, res), _job.expected_duration))
        
    def job_draw(self, i):
        running_jobs = self.datasource['running_jobs']
        current_time = self.datasource['current_time']
        if current_time is None:
            self.stop()
            return
        self.xlimit(current_time)
        if not running_jobs:
            _finished = list(self.job_patches.keys())
            if _finished:
                self.update_finished_jobs(current_time, _finished)
            for res, ax in self.axis_mapper.items():
                self.modify_legend(ax, res) 
                self.update_ylabel(ax, res, 0)            
            
            return 
            
        _resources = self.system_info.keys()
        self.start_line_draw(current_time)
        
        # Load new jobs 
        self.load_new_jobs(current_time, running_jobs, _resources)
                
        # Remove jobs already finished
        _finished = [_id for _id in self.job_patches.keys() if _id not in running_jobs ]  # filter(lambda x: x not in running_jobs, list(self.job_patches.keys()))
        if _finished:
            self.update_finished_jobs(current_time, _finished)
            
        # Create new Rectangles or update its position (By increasing consumption)
        _stop_f = False
        for res in _resources:
            ax = self.axis_mapper[res]
            # Sorted by consumption
            _sorted_jobs = self.sorted_jobs[res].get_reversed_object_list()
            # Start from 0
            self.system_info[res]['used'] = 0
            
            for sjob in _sorted_jobs:
                _id = sjob.id
                job = self.job_object[_id]
                j_used = getattr(sjob, res)
                if j_used == 0:
                    # If 0 does not exists the patch
                    continue
                # Check if already exists
                if _id in self.job_patches:
                    _patches = self.job_patches[_id]
                    # Check if the patch exists
                    if res in _patches:
                        self.update_running_job(_patches, res, j_used)
                        continue
                else:
                    self.job_patches[_id] = {}
                    self.job_color[_id] = self.next_color()                
                
                a_used = self.system_info[res]['used']  # / self.system_info[res]['total']
                # _min, _max = self.job_normalization(current_time, sjob)
                r = patches.Rectangle(
                    (job.start_time, a_used / self.system_info[res]['total']), sjob.expected_duration, j_used / self.system_info[res]['total'],
                    fill=True, linewidth=0, edgecolor='black', facecolor=self.job_color[_id], linestyle='dashed', label=_id
                )                 
                self.job_patches[_id][res] = r 
                self.system_info[res]['used'] += j_used
                ax.add_patch(r)        
                 
            self.modify_legend(ax, res)
            self.update_ylabel(ax, res, self.system_info[res]['used'])
        plt.rcParams.update({'font.size': 16})
        self.figure.canvas.draw()
        self.figure.canvas.flush_events()             
        
    def update_ylabel(self, _ax, _res, _used):
        _res_usage = _used / self.system_info[_res]['total']
        _ax.set_ylabel('{} {:.2%}'.format(_res, _res_usage), rotation=90, size=20)
        
    def modify_legend(self, ax, res, legend_size=10):
        h, l = ax.get_legend_handles_labels()
        # L.get_texts()[0].set_text('make it short')
        if len(l) <= legend_size:
            ax.legend(h, l, prop={'size':10}, loc='upper right', title='Job id')
            return
        _generated_legend = {l: h for h, l  in zip(h, l)}
        _modified_legend = []
        for _idx, _job in enumerate(self.sorted_jobs[res].get_reversed_object_list()):
            if _idx == legend_size:
                break
            _modified_legend.append((str(_job.id), _generated_legend[str(_job.id)]))
        l, h = zip(*_modified_legend)
        ax.legend(h, l, prop={'size':12}, loc='upper left')
        
    def update_running_job(self, _patches, _res, _used):
        _patch = _patches[_res]
        _patch.set_y(self.system_info[_res]['used'] / self.system_info[_res]['total'])
        self.system_info[_res]['used'] += _used 
        # _patch.set_height(self.system_info[_res]['used'] / self.system_info[_res]['total'])
        
    def update_finished_jobs(self, current_time, _finished_jobs):
        for _id in _finished_jobs:
            self.remove_job(_id)
        
    def remove_job(self, id):
        _dict_patches = self.job_patches.pop(id, {})
        
        for res, _list in self.sorted_jobs.items(): 
            _job = _list.pop(id=id)
            self.system_info[res]['used'] -= getattr(_job, res)
            if res in _dict_patches:
                _dict_patches[res].remove()

        del self.job_color[id]
        del self.job_object[id]
        
    def job_normalization(self, actual_time, job):
        _min = self.min_time
        _max = self.max_time
        _job_start = (actual_time - job.start_time)
        _job_end = (job.start_time + job.expected_duration - actual_time)
        if _job_start > _min:
             _min = _job_start
        if _job_end < _max:
            _max = _job_end
        return (_min, _max)
    
    def start_line_draw(self, actual_time):
        if self.last_line['time'] == actual_time:
            return
        self.last_line['time'] = actual_time
        _text = self.last_line.pop('text', None)
        if _text is not None:
            _text.remove()
        self.last_line['text'] = self.figure.text(0.62, 0.05, 'Actual time: {}'.format(str_datetime(actual_time)), fontsize=12, backgroundcolor='white')
        for name, ax in self.axis_mapper.items():
            _old = self.last_line.pop(name, None)
            if _old is not None:
                _old.remove()
            self.last_line[name] = ax.axvline(x=actual_time, color='red')
        
            
    def create_empty_map(self):
        sys_config = self.system_info
        for _k in sys_config:
            sys_config[_k]['used'] = 0
        types = len(sys_config)
        self.figure, axis = plt.subplots(types, 1, sharex=True, sharey=True, figsize=(10, 8))
        sys_plot = {}
        for (ax, _sys) in zip(axis, sys_config):
            self.xlimit(0, ax)
            ax.set_ylabel(_sys, rotation=90, size=12)  # , size='large')
            ax.grid()
            xticks = ax.yaxis.get_major_ticks() 
            xticks[0].label1.set_visible(False)
            xticks[-1].label1.set_visible(False)
            sys_plot[_sys] = ax            

            for label in (ax.get_xticklabels() + ax.get_yticklabels()):
                label.set_fontsize(16)
        self.axis_mapper = sys_plot
        self.figure.subplots_adjust(hspace=0)
        plt.setp([a.get_xticklabels() for a in self.figure.axes[:-1]], visible=False)
        self.animation = Animation.FuncAnimation(self.figure, self.job_draw, interval=1000)
        
        plt.rcParams.update({'font.size': 16})
        plt.show()
                    
    def xlimit(self, actual_time, ax=None):
        if ax is None:         
            for _, ax in self.axis_mapper.items():
                self._set_xlimit(ax, actual_time)
        else: 
            self._set_xlimit(ax, actual_time)
            
    def _set_xlimit(self, ax, actual_time):
        if not actual_time:
            return
        self.xlim = (actual_time - self.min_time, actual_time + self.max_time)
        ax.set_xlim(self.xlim)
        labels = ['-600', '-{}'.format(actual_time % 600), '+600', '+1200', '+1800' ]
        ax.set_xticklabels(labels)
            
    
    def next_color(self):
        if self.color_pos is None: 
            self.color_pos = len(self.colors) // 2
        while True:
            _next_pos = self.color_pos
            self.color_pos += 1
            if self.color_pos == len(self.colors):
                self.color_pos = 0 
            if self.colors[_next_pos] not in self.color_blacklist:
                break
        return self.colors[_next_pos] 
        
    def set_colors(self):
        self.colors = [name for name, hex in colors.cnames.items()]
